from collections import defaultdict
from typing import Dict, List

import click
from beaker import Beaker, Cluster, Node, Priority
from rich import print
from rich.progress import Progress
from rich.table import Table

from .main import CLICK_COMMAND_DEFAULTS, CLICK_GROUP_DEFAULTS, main


@main.group(**CLICK_GROUP_DEFAULTS)
def cluster():
    """
    Get information on Beaker clusters.
    """


@cluster.command(name="list", **CLICK_COMMAND_DEFAULTS)
@click.option(
    "--cloud",
    is_flag=True,
    help="""Only show cloud clusters.""",
)
def list_clusters(cloud: bool = False):
    """
    List available clusters.

    By default only on-premise clusters are displayed.
    """
    beaker = Beaker.from_env(session=True)

    table = Table(title="Clusters", show_lines=True)
    table.add_column("Cluster", justify="left", no_wrap=True)
    table.add_column("Nodes")

    def cluster_info(cluster: Cluster) -> str:
        info = f"{icon} [b magenta]{cluster.full_name}[/], {len(nodes)} nodes"
        if (limits := cluster.node_spec) is not None:
            info += f"\nCPUs: {limits.cpu_count}\n"
            info += f"GPUs: {limits.gpu_count or 0} {'x' if limits.gpu_type else ''} {limits.gpu_type or ''}"
        return info

    def node_info(nodes: List[Node]) -> str:
        return "\n".join(
            f"[i cyan]{node.hostname}[/] - "
            f"CPUs: {node.limits.cpu_count}, "
            f"GPUs: {node.limits.gpu_count or 0} {'x' if node.limits.gpu_type else ''} {node.limits.gpu_type or ''}"
            for node in nodes
        )

    with Progress(transient=True) as progress:
        task = progress.add_task("Collecting clusters...", start=False, total=None)
        clusters = [c for c in beaker.cluster.list() if c.is_cloud == cloud]
        progress.update(task, completed=True)

    for cluster in clusters:
        icon = "☁️" if cluster.is_cloud else "🏠"
        nodes = sorted(beaker.cluster.nodes(cluster), key=lambda node: node.hostname)
        table.add_row(
            cluster_info(cluster),
            node_info(nodes),
        )

    print(table)


def complete_cluster_name(ctx, param, incomplete) -> List[str]:
    del ctx, param
    beaker = Beaker.from_env()
    return [c.full_name for c in beaker.cluster.list() if c.full_name.startswith(incomplete)]


@cluster.command(name="util", **CLICK_COMMAND_DEFAULTS)
@click.argument(
    "cluster_name", nargs=1, required=True, type=str, shell_complete=complete_cluster_name
)
@click.option("--nodes", is_flag=True, help="Show details of each node.")
def cluster_util(cluster_name: str, nodes: bool = False):
    """
    Get the current status and utilization for a cluster.
    """
    beaker = Beaker.from_env(session=True)

    with Progress(transient=True) as progress:
        task = progress.add_task("Pulling cluster data...", start=False, total=None)
        cluster_util = beaker.cluster.utilization(cluster_name)
        cluster = cluster_util.cluster
        icon = "☁️" if cluster.is_cloud else "🏠"
        progress.update(task, completed=True)

    free_nodes = 0
    total_nodes = len(cluster_util.nodes)
    cordoned_nodes = 0
    free_gpus = 0
    total_gpus = 0
    for node_util in cluster_util.nodes:
        total_gpus += node_util.limits.gpu_count or 0
        if not node_util.cordoned:
            if node_util.running_jobs == 0:
                free_nodes += 1
            free_gpus += node_util.free.gpu_count or 0
        else:
            cordoned_nodes += 1

    running_jobs_by_priority: Dict[Priority, int] = defaultdict(int)
    running_preemptible_jobs_by_priority: Dict[Priority, int] = defaultdict(int)
    queued_jobs_by_priority: Dict[Priority, int] = defaultdict(int)
    for job in cluster_util.jobs:
        if job.priority is None:
            continue
        if job.is_running:
            running_jobs_by_priority[job.priority] += 1
            if job.is_preemptible:
                running_preemptible_jobs_by_priority[job.priority] += 1
        elif job.is_queued:
            queued_jobs_by_priority[job.priority] += 1

    def get_node_notes() -> str:
        notes = []
        if cluster.require_preemptible_tasks:
            notes.append("[yellow]All tasks must be preemptible.[/]")
        if cordoned_nodes == 1:
            notes.append(f"[yellow]{cordoned_nodes} node is cordoned.[/]")
        elif cordoned_nodes > 1:
            notes.append(f"[yellow]{cordoned_nodes} nodes are cordoned.[/]")
        return "\n".join(notes)

    summary_table = Table(
        title=f"{icon} [b magenta]{cluster.full_name}[/]\nSummary",
        show_lines=True,
        show_header=False,
    )
    summary_table.add_column("Name", justify="left", no_wrap=True)
    summary_table.add_column("Value")
    summary_table.add_row("[cyan]URL[/]", f"[i u blue]{beaker.cluster.url(cluster)}[/]")
    summary_table.add_row("[cyan]Notes[/]", get_node_notes())
    summary_table.add_row(
        "[cyan]Running jobs[/]",
        f"[bold]{cluster_util.running_jobs} total ({cluster_util.running_preemptible_jobs} preemptible)[/]\n"
        "------------------------------------------\n"
        f"[red bold i]Urgent priority:.....[/] {running_jobs_by_priority[Priority.urgent]} ({running_preemptible_jobs_by_priority[Priority.urgent]} preemptible)\n"
        f"[red i]High priority:.......[/] {running_jobs_by_priority[Priority.high]} ({running_preemptible_jobs_by_priority[Priority.high]} preemptible)\n"
        f"[green i]Normal priority:.....[/] {running_jobs_by_priority[Priority.normal]} ({running_preemptible_jobs_by_priority[Priority.normal]} preemptible)\n"
        f"[cyan i]Low priority:........[/] {running_jobs_by_priority[Priority.low]} ({running_preemptible_jobs_by_priority[Priority.low]} preemptible)\n"
        f"[blue i]Preemptible priority:[/] {running_jobs_by_priority[Priority.preemptible]}",
    )
    summary_table.add_row(
        "[cyan]Queued jobs[/]",
        f"[bold]{cluster_util.queued_jobs} total[/]\n"
        "------------------------------------------\n"
        f"[red bold i]Urgent priority:.....[/] {queued_jobs_by_priority[Priority.urgent]}\n"
        f"[red i]High priority:.......[/] {queued_jobs_by_priority[Priority.high]}\n"
        f"[green i]Normal priority:.....[/] {queued_jobs_by_priority[Priority.normal]}\n"
        f"[cyan i]Low priority:........[/] {queued_jobs_by_priority[Priority.low]}\n"
        f"[blue i]Preemptible priority:[/] {queued_jobs_by_priority[Priority.preemptible]}",
    )
    summary_table.add_row(
        "[cyan]Free nodes[/]", f"[{'green' if free_nodes else 'red'}]{free_nodes}/{total_nodes}[/]"
    )
    summary_table.add_row(
        "[cyan]Free GPUs[/]", f"[{'green' if free_gpus else 'red'}]{free_gpus}/{total_gpus}[/]"
    )

    print(summary_table)

    if nodes:
        node_table = Table(
            title=f"{icon} [b magenta]{cluster.full_name}[/]\nNodes",
            show_lines=True,
        )
        node_table.add_column("Node", justify="left", no_wrap=True)
        node_table.add_column("Jobs")
        node_table.add_column("Utilization")

        for node_util in sorted(cluster_util.nodes, key=lambda n: n.hostname):
            node_table.add_row(
                f"[i cyan]{node_util.hostname}[/]",
                f"{node_util.running_jobs} jobs ({node_util.running_preemptible_jobs} preemptible)",
                "[red]\N{ballot x} cordoned[/]"
                if node_util.cordoned
                else f"CPUs free: [{'green' if node_util.free.cpu_count else 'red'}]"
                f"{node_util.free.cpu_count} / {node_util.limits.cpu_count}[/]\n"
                f"GPUs free: [{'green' if node_util.free.gpu_count else 'red'}]"
                f"{node_util.free.gpu_count or 0} / {node_util.limits.gpu_count}[/] {node_util.free.gpu_type or ''}",
            )

        print(node_table)



from typing import List, Optional, Sequence

import click
from beaker import Beaker, Experiment, ExperimentConflict, ExperimentNotFound
from rich import print, prompt

from .. import util
from ..exceptions import ConfigurationError, NotFoundError
from .main import CLICK_COMMAND_DEFAULTS, main


@main.command(**CLICK_COMMAND_DEFAULTS)
@click.argument("experiment", nargs=-1, type=str)
@click.option("-l", "--latest", is_flag=True, help="""Stop your latest experiment.""")
@click.option(
    "-w",
    "--workspace",
    type=str,
    help="""The Beaker workspace to pull experiments from.
    If not specified, your default workspace will be used.""",
)
@click.option("--dry-run", is_flag=True, help="Do a dry-run without stopping any experiments.")
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    help="""Skip all confirmation prompts.""",
)
def stop(
    experiment: Sequence[str],
    latest: bool = False,
    workspace: Optional[str] = None,
    dry_run: bool = False,
    yes: bool = False,
):
    """
    Stop a running experiment.
    """
    if experiment and latest:
        raise ConfigurationError("-l/--latest is mutually exclusive with [EXPERIMENT] args")

    beaker = Beaker.from_env(session=True)

    experiments: List[Experiment] = []
    if experiment:
        for experiment_name in experiment:
            try:
                experiments.append(beaker.experiment.get(experiment_name))
            except ExperimentNotFound:
                raise NotFoundError(f"Experiment '{experiment_name}' not found")
    elif latest:
        exp = util.get_latest_experiment(beaker, workspace=workspace, running=True)
        if exp is None:
            print("[yellow]No running experiments to stop[/]")
        else:
            experiments.append(exp)

    for exp in experiments:
        if dry_run:
            print(f"[b yellow]Dry run:[/] would stop [b cyan]{exp.name}[/]")
        else:
            if not yes and not prompt.Confirm.ask(
                f"Stop experiment [b cyan]{exp.name}[/] at [blue u]{beaker.experiment.url(exp)}[/]?"
            ):
                print("[yellow]Skipping experiment...[/]")
                continue

            try:
                beaker.experiment.stop(exp)
            except (ExperimentNotFound, ExperimentConflict):
                # Beaker API may return 404 if the experiment was already canceled.
                pass
            print(
                f"[b green]\N{check mark}[/] [b cyan]{exp.name}[/] at {beaker.experiment.url(exp)} stopped"
            )



import json
import tempfile
import time
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import Optional, Tuple, cast

import requests
import rich
from beaker import (
    Beaker,
    Dataset,
    DatasetConflict,
    DatasetNotFound,
    Digest,
    Experiment,
    Job,
    JobKind,
    SecretNotFound,
    WorkspaceNotSet,
)
from rich import print, prompt
from rich.console import Console

from . import constants
from .constants import GITHUB_TOKEN_SECRET
from .exceptions import *
from .version import VERSION

VERSION_CHECK_INTERVAL = 12 * 3600  # 12 hours
DEFAULT_INTERNAL_CONFIG_LOCATION: Optional[Path] = None
try:
    DEFAULT_INTERNAL_CONFIG_LOCATION = Path.home() / ".beaker" / ".beaker-gantry.json"
except RuntimeError:
    # Can't locate home directory.
    pass


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return self.value


@dataclass
class InternalConfig:
    version_checked: Optional[float] = None

    @classmethod
    def load(cls) -> Optional["InternalConfig"]:
        path = DEFAULT_INTERNAL_CONFIG_LOCATION
        if path is None:
            return None
        elif path.is_file():
            with open(path, "r") as f:
                return cls(**json.load(f))
        else:
            return cls()

    def save(self):
        path = DEFAULT_INTERNAL_CONFIG_LOCATION
        if path is None:
            return None
        else:
            path.parent.mkdir(exist_ok=True, parents=True)
            with open(path, "w") as f:
                json.dump(asdict(self), f)


def unique_name() -> str:
    import uuid

    import petname

    return cast(str, petname.generate()) + "-" + str(uuid.uuid4())[:7]


def stderr_console() -> Console:
    return Console(stderr=True)


def print_stderr(*args, **kwargs):
    stderr_console().print(*args, **kwargs)


def print_exception(*args, **kwargs):
    stderr_console().print_exception(*args, **kwargs)


def parse_git_remote_url(url: str) -> Tuple[str, str]:
    """
    Parse a git remote URL into a GitHub (account, repo) pair.

    :raises InvalidRemoteError: If the URL can't be parsed correctly.
    """
    try:
        account, repo = (
            url.split("https://github.com/")[-1]
            .split("git@github.com:")[-1]
            .split(".git")[0]
            .split("/")
        )
    except ValueError:
        raise InvalidRemoteError(f"Failed to parse GitHub repo path from remote '{url}'")
    return account, repo


def get_latest_experiment(
    beaker: Beaker,
    *,
    author: Optional[str] = None,
    workspace: Optional[str] = None,
    running: bool = False,
) -> Optional[Experiment]:
    workspace_id = beaker.workspace.get(workspace=workspace).id

    jobs = [
        job
        for job in beaker.job.list(
            author=author if author is not None else beaker.account.whoami(),
            kind=JobKind.execution,
            finalized=False,
        )
        if job.workspace == workspace_id
    ]

    if running:
        jobs = [job for job in jobs if job.is_running]
        jobs = sorted(
            jobs,
            key=lambda j: j.status.started,  # type: ignore
            reverse=True,
        )
    else:
        jobs = sorted(jobs, key=lambda j: j.status.created, reverse=True)

    if jobs:
        job = jobs[0]
        assert job.execution is not None
        return beaker.experiment.get(job.execution.experiment)
    else:
        return None


def follow_experiment(
    beaker: Beaker, experiment: Experiment, timeout: int = 0, tail: bool = False
) -> Job:
    console = rich.get_console()

    # Wait for job to start...
    job: Optional[Job] = beaker.experiment.tasks(experiment.id)[0].latest_job  # type: ignore
    if job is None:
        print("Waiting for job to launch..", end="")
        while job is None:
            time.sleep(1.0)
            print(".", end="")
            job = beaker.experiment.tasks(experiment.id)[0].latest_job  # type: ignore

    exit_code: Optional[int] = job.status.exit_code

    stream_logs = exit_code is None and not job.is_finalized
    if stream_logs:
        print()
        rich.get_console().rule("Logs")
        for line_bytes in beaker.job.follow(
            job,
            timeout=timeout if timeout > 0 else None,
            include_timestamps=False,
            since=datetime.utcnow() - timedelta(seconds=5) if tail else None,
        ):
            line = line_bytes.decode(errors="ignore")
            if line.endswith("\n"):
                line = line[:-1]
            console.print(line, highlight=False, markup=False)
        rich.get_console().rule("End logs")
        print()

        # Refresh the job.
        job = beaker.job.get(job.id)

    return job


def display_logs(beaker: Beaker, job: Job) -> Job:
    console = rich.get_console()
    print()
    rich.get_console().rule("Logs")
    for line_bytes in beaker.job.logs(job, quiet=True):
        line = line_bytes.decode(errors="ignore")
        if line.endswith("\n"):
            line = line[:-1]
        console.print(line, highlight=False, markup=False)
    rich.get_console().rule("End logs")
    print()

    # Refresh the job.
    job = beaker.job.get(job.id)
    return job


def display_results(beaker: Beaker, experiment: Experiment, job: Job):
    exit_code = job.status.exit_code
    if exit_code is None:
        raise ExperimentFailedError(
            f"Experiment failed, see {beaker.experiment.url(experiment)} for details"
        )
    elif exit_code > 0:
        raise ExperimentFailedError(
            f"Experiment exited with non-zero code ({exit_code}), see {beaker.experiment.url(experiment)} for details"
        )
    assert job.execution is not None
    assert job.status.started is not None
    assert job.status.exited is not None
    result_dataset = None
    if job.result is not None and job.result.beaker is not None:
        result_dataset = job.result.beaker

    print(
        f"[b green]\N{check mark}[/] [b cyan]{experiment.name}[/] completed successfully\n"
        f"[b]Experiment:[/] {beaker.experiment.url(experiment)}\n"
        f"[b]Runtime:[/] {format_timedelta(job.status.exited - job.status.started)}\n"
        f"[b]Results:[/] {None if result_dataset is None else beaker.dataset.url(result_dataset)}"
    )

    metrics = beaker.experiment.metrics(experiment, task=job.execution.task)
    if metrics is not None:
        print("[b]Metrics:[/]", metrics)


def ensure_repo(allow_dirty: bool = False) -> Tuple[str, str, str, bool]:
    from git.repo import Repo

    repo = Repo(".")

    # Parse account name, repo name, and current commit.
    account, repo_name = parse_git_remote_url(repo.remote().url)
    git_ref = str(repo.commit())

    # Check if repo is dirty (uncommitted changes).
    if repo.is_dirty() and not allow_dirty:
        raise DirtyRepoError("You have uncommitted changes! Use --allow-dirty to force.")

    # Check if repo is public.
    response = requests.get(f"https://github.com/{account}/{repo_name}")
    if response.status_code not in {200, 404}:
        response.raise_for_status()
    is_public = response.status_code == 200

    return account, repo_name, git_ref, is_public


def ref_exists_on_remote(git_ref: str) -> bool:
    from git.cmd import Git

    git = Git(".")

    output = cast(
        str, git.execute(["git", "branch", "-r", "--contains", git_ref], stdout_as_string=True)
    )
    output = output.strip()
    return len(output) > 0


def ensure_entrypoint_dataset(beaker: Beaker) -> Dataset:
    import hashlib
    from importlib.resources import read_binary

    import gantry

    workspace_id = beaker.workspace.get().id

    # Get hash of the local entrypoint source file.
    sha256_hash = hashlib.sha256()
    contents = replace_tags(read_binary(gantry, constants.ENTRYPOINT))
    sha256_hash.update(contents)

    entrypoint_dataset_name = f"gantry-v{VERSION}-{workspace_id}-{sha256_hash.hexdigest()[:6]}"

    # Ensure gantry entrypoint dataset exists.
    gantry_entrypoint_dataset: Dataset
    try:
        gantry_entrypoint_dataset = beaker.dataset.get(entrypoint_dataset_name)
    except DatasetNotFound:
        # Create it.
        print(f"Creating entrypoint dataset '{entrypoint_dataset_name}'")
        try:
            with tempfile.TemporaryDirectory() as tmpdirname:
                tmpdir = Path(tmpdirname)
                entrypoint_path = tmpdir / constants.ENTRYPOINT
                with open(entrypoint_path, "wb") as entrypoint_file:
                    entrypoint_file.write(contents)
                gantry_entrypoint_dataset = beaker.dataset.create(
                    entrypoint_dataset_name, entrypoint_path
                )
        except DatasetConflict:  # could be in a race with another `gantry` process.
            time.sleep(1.0)
            gantry_entrypoint_dataset = beaker.dataset.get(entrypoint_dataset_name)

    # Verify contents.
    ds_files = list(beaker.dataset.ls(gantry_entrypoint_dataset))
    for retry in range(1, 4):
        ds_files = list(beaker.dataset.ls(gantry_entrypoint_dataset))
        if len(ds_files) >= 1:
            break
        else:
            time.sleep(1.5**retry)

    if len(ds_files) != 1:
        raise EntrypointChecksumError(
            f"Entrypoint dataset {beaker.dataset.url(gantry_entrypoint_dataset)} is missing the "
            f"required entrypoint file. Please run again."
        )

    if ds_files[0].digest != Digest.from_decoded(sha256_hash.digest(), "SHA256"):
        raise EntrypointChecksumError(
            f"Checksum failed for entrypoint dataset {beaker.dataset.url(gantry_entrypoint_dataset)}\n"
            f"This could be a bug, or it could mean someone has tampered with the dataset.\n"
            f"If you're sure no one has tampered with it, you can delete the dataset from "
            f"the Beaker dashboard and try again."
        )

    return gantry_entrypoint_dataset


def ensure_github_token_secret(
    beaker: Beaker, secret_name: str = constants.GITHUB_TOKEN_SECRET
) -> str:
    try:
        beaker.secret.get(secret_name)
    except SecretNotFound:
        raise GitHubTokenSecretNotFound(
            f"GitHub token secret '{secret_name}' not found in Beaker workspace!\n"
            f"You can create a suitable GitHub token by going to https://github.com/settings/tokens/new "
            f"and generating a token with '\N{ballot box with check} repo' scope.\n"
            f"Then upload your token as a Beaker secret using the Beaker CLI or Python client."
        )
    return secret_name


def format_timedelta(td: "timedelta") -> str:
    def format_value_and_unit(value: int, unit: str) -> str:
        if value == 1:
            return f"{value} {unit}"
        else:
            return f"{value} {unit}s"

    parts = []
    seconds = int(td.total_seconds())
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if days:
        parts.append(format_value_and_unit(days, "day"))
    if hours:
        parts.append(format_value_and_unit(hours, "hour"))
    if minutes:
        parts.append(format_value_and_unit(minutes, "minute"))
    if seconds:
        parts.append(format_value_and_unit(seconds, "second"))
    return ", ".join(parts)


def check_for_upgrades():
    config = InternalConfig.load()
    if (
        config is not None
        and config.version_checked is not None
        and (time.time() - config.version_checked <= VERSION_CHECK_INTERVAL)
    ):
        return

    import packaging.version
    import requests

    try:
        response = requests.get(
            "https://api.github.com/repos/allenai/beaker-gantry/releases/latest", timeout=1
        )
        if response.ok:
            latest_version = packaging.version.parse(response.json()["tag_name"])
            if latest_version > packaging.version.parse(VERSION):
                print_stderr(
                    f":warning: [yellow]You're using [b]gantry v{VERSION}[/], "
                    f"but a newer version ([b]v{latest_version}[/]) is available: "
                    f"https://github.com/allenai/beaker-gantry/releases/tag/v{latest_version}[/]\n"
                    f"[yellow i]You can upgrade by running:[/] pip install --upgrade beaker-gantry beaker-py\n",
                )
            if config is not None:
                config.version_checked = time.time()
                config.save()
    except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
        pass


def ensure_workspace(
    workspace: Optional[str] = None,
    yes: bool = False,
    gh_token_secret: str = GITHUB_TOKEN_SECRET,
    public_repo: bool = False,
) -> Beaker:
    beaker = (
        Beaker.from_env(session=True)
        if workspace is None
        else Beaker.from_env(session=True, default_workspace=workspace)
    )
    try:
        permissions = beaker.workspace.get_permissions()
        if (
            not public_repo
            and permissions.authorizations is not None
            # a default user called "ai2" gets added to every workspace
            and len(permissions.authorizations) > 2
        ):
            print_stderr(
                f"[yellow]Your workspace [b]{beaker.workspace.url()}[/] has multiple contributors! "
                f"Every contributor can view your GitHub personal access token secret ('{gh_token_secret}').[/]"
            )
            if not yes and not prompt.Confirm.ask(
                "[yellow][i]Are you sure you want to use this workspace?[/][/]"
            ):
                raise KeyboardInterrupt
        elif workspace is None:
            default_workspace = beaker.workspace.get()
            if not yes and not prompt.Confirm.ask(
                f"Using default workspace [b cyan]{default_workspace.full_name}[/]. [i]Is that correct?[/]"
            ):
                raise KeyboardInterrupt
    except WorkspaceNotSet:
        raise ConfigurationError(
            "'--workspace' option is required since you don't have a default workspace set"
        )
    return beaker


def replace_tags(contents: bytes) -> bytes:
    tag_start = contents.find(b"${{")
    while tag_start != -1:
        tag_end = contents.find(b"}}") + 2
        tag = contents[tag_start:tag_end]
        constant_name = tag.split(b" ")[1].decode()
        contents = contents.replace(tag, getattr(constants, constant_name).encode())  # type: ignore
        tag_start = contents.find(b"${{", tag_end)
    assert b"${{" not in contents
    return contents



class GantryError(Exception):
    """
    Base exception for all error types that Gantry might raise.
    """


class GitError(GantryError):
    pass


class DirtyRepoError(GitError):
    pass


class UnpushedChangesError(GitError):
    pass


class InvalidRemoteError(GitError):
    pass


class ConfigurationError(GantryError):
    pass


class ExperimentFailedError(GantryError):
    pass


class EntrypointChecksumError(GantryError):
    pass


class GitHubTokenSecretNotFound(GantryError):
    pass


class TermInterrupt(GantryError):
    pass


class NotFoundError(GantryError):
    pass



import platform
import random
import string
import sys
from fnmatch import fnmatch
from pathlib import Path
from typing import List, Optional, Tuple, Union

import click
import rich
from beaker import (
    Beaker,
    ExperimentConflict,
    ExperimentSpec,
    ImageNotFound,
    Job,
    JobTimeoutError,
    Priority,
    RetrySpec,
    SecretNotFound,
    TaskResources,
    TaskSpec,
)
from rich import print, prompt

from .. import constants, util
from ..aliases import PathOrStr
from ..exceptions import *
from ..util import print_stderr
from ..version import VERSION
from .main import CLICK_COMMAND_DEFAULTS, main


@main.command(**CLICK_COMMAND_DEFAULTS)
@click.argument("arg", nargs=-1)
@click.option(
    "-n",
    "--name",
    type=str,
    help="""A name to assign to the experiment on Beaker. Defaults to a randomly generated name.""",
)
@click.option(
    "-t",
    "--task-name",
    type=str,
    help="""A name to assign to the task on Beaker.""",
    default="main",
    show_default=True,
)
@click.option("-d", "--description", type=str, help="""A description for the experiment.""")
@click.option(
    "-w",
    "--workspace",
    type=str,
    help="""The Beaker workspace to use.
    If not specified, your default workspace will be used.""",
)
@click.option(
    "-c",
    "--cluster",
    type=str,
    multiple=True,
    default=None,
    help="""A potential cluster to use. This option can be used multiple times to allow multiple clusters.
    You also specify it as a wildcard, e.g. '--cluster ai2/*-cirrascale'.
    If you don't specify a cluster or the priority, the priority will default to 'preemptible' and
    the job will be able to run on any on-premise cluster.""",
    show_default=True,
)
@click.option(
    "--hostname",
    type=str,
    multiple=True,
    default=None,
    help="""Hostname constraints to apply to the experiment spec. This option can be used multiple times to allow
    multiple hosts.""",
    show_default=True,
)
@click.option(
    "--beaker-image",
    type=str,
    help=f"""The name or ID of an image on Beaker to use for your experiment.
    Mutually exclusive with --docker-image. Defaults to '{constants.DEFAULT_IMAGE}' if neither is set.""",
)
@click.option(
    "--docker-image",
    type=str,
    help="""The name of a public Docker image to use for your experiment.
    Mutually exclusive with --beaker-image.""",
)
@click.option(
    "--cpus",
    type=float,
    help="""Minimum number of logical CPU cores (e.g. 4.0, 0.5).""",
)
@click.option(
    "--gpus",
    type=int,
    help="""Minimum number of GPUs (e.g. 1).""",
)
@click.option(
    "--memory",
    type=str,
    help="""Minimum available system memory as a number with unit suffix (e.g. 2.5GiB).""",
)
@click.option(
    "--shared-memory",
    type=str,
    help="""Size of /dev/shm as a number with unit suffix (e.g. 2.5GiB).""",
)
@click.option(
    "--dataset",
    type=str,
    multiple=True,
    help="""An input dataset in the form of 'dataset-name:/mount/location' or
    'dataset-name:sub/path:/mount/location' to attach to your experiment.
    You can specify this option more than once to attach multiple datasets.""",
)
@click.option(
    "--gh-token-secret",
    type=str,
    help="""The name of the Beaker secret that contains your GitHub token.""",
    default=constants.GITHUB_TOKEN_SECRET,
    show_default=True,
)
@click.option(
    "--ref",
    type=str,
    help="""The target git ref to use. This defaults to the current commit.""",
)
@click.option(
    "--conda",
    type=click.Path(exists=True, dir_okay=False),
    help=f"""Path to a conda environment file for reconstructing your Python environment.
    If not specified, '{constants.CONDA_ENV_FILE}' will be used if it exists.""",
)
@click.option(
    "--pip",
    type=click.Path(exists=True, dir_okay=False),
    help=f"""Path to a PIP requirements file for reconstructing your Python environment.
    If not specified, '{constants.PIP_REQUIREMENTS_FILE}' will be used if it exists.""",
)
@click.option(
    "--venv",
    type=str,
    help="""The name of an existing conda environment on the image to use.""",
)
@click.option(
    "--env",
    type=str,
    help="""Environment variables to add the Beaker experiment. Should be in the form '{NAME}={VALUE}'.""",
    multiple=True,
)
@click.option(
    "--env-secret",
    type=str,
    help="""Environment variables to add the Beaker experiment from Beaker secrets.
    Should be in the form '{NAME}={SECRET_NAME}'.""",
    multiple=True,
)
@click.option(
    "--dataset-secret",
    type=str,
    help="""Mount a Beaker secret to a file as a dataset.
    Should be in the form '{SECRET_NAME}:{MOUNT_PATH}'.""",
    multiple=True,
)
@click.option(
    "--show-logs/--no-logs",
    default=True,
    show_default=True,
    help="""Whether or not to stream the logs to stdout as the experiment runs.
    This only takes effect when --timeout is non-zero.""",
)
@click.option(
    "--timeout",
    type=int,
    default=0,
    help="""Time to wait (in seconds) for the experiment to finish.
    A timeout of -1 means wait indefinitely.
    A timeout of 0 means don't wait at all.""",
    show_default=True,
)
@click.option(
    "--allow-dirty",
    is_flag=True,
    help="""Allow submitting the experiment with a dirty working directory.""",
)
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    help="""Skip all confirmation prompts.""",
)
@click.option("--dry-run", is_flag=True, help="""Do a dry run only.""")
@click.option(
    "--save-spec",
    type=click.Path(dir_okay=False, file_okay=True),
    help="""A path to save the generated Beaker experiment spec to.""",
)
@click.option(
    "--priority",
    type=click.Choice([str(p) for p in Priority]),
    help="The job priority. If you don't specify at least one cluster, priority will default to 'preemptible'.",
)
@click.option(
    "--install",
    type=str,
    help="""Override the default installation command, e.g. '--install "python setup.py install"'""",
)
@click.option(
    "--no-python",
    is_flag=True,
    help="""If set, gantry will skip setting up a Python environment altogether.""",
)
@click.option(
    "--no-conda",
    is_flag=True,
    help="""If set, gantry will not use conda to construct a Python environment,
    and instead will potentially use the default Python environment on the image.""",
)
@click.option(
    "--replicas",
    type=int,
    help="""The number of task replicas to run.""",
)
@click.option(
    "--leader-selection",
    is_flag=True,
    help="""Specifies that the first task replica should be the leader and populates each task
    with 'BEAKER_LEADER_REPLICA_HOSTNAME' and 'BEAKER_LEADER_REPLICA_NODE_ID' environment variables.
    This is only applicable when '--replicas INT' and '--host-networking' are used,
    although the '--host-networking' flag can be omitted in this case since it's assumed.""",
)
@click.option(
    "--host-networking",
    is_flag=True,
    help="""Specifies that each task replica should use the host's network.
    When used with '--replicas INT', this allows the replicas to communicate with each
    other using their hostnames.""",
)
@click.option(
    "--propagate-failure", is_flag=True, help="""Stop the experiment if any task fails."""
)
@click.option(
    "--propagate-preemption", is_flag=True, help="""Stop the experiment if any task is preempted."""
)
@click.option(
    "--synchronized-start-timeout",
    type=str,
    help="""
    If set, jobs in the replicated task will wait this long to start until all other jobs are also ready.
    """,
)
@click.option(
    "-m",
    "--mount",
    type=str,
    help="""Host directories to mount to the Beaker experiment. Should be in the form '{HOST_SOURCE}:{TARGET}'
    similar to the '-v' option with 'docker run'.""",
    multiple=True,
)
@click.option(
    "--weka",
    type=str,
    multiple=True,
    help="""A weka bucket to mount in the form of 'bucket-name:/mount/location',
    e.g. --weka=oe-training-default:/data""",
)
@click.option(
    "-b", "--budget", type=str, help="""The budget account to associate with the experiment."""
)
@click.option("--preemptible", is_flag=True, help="""Mark the job as preemptible.""")
@click.option("--stop-preemptible", is_flag=True, help="""Stop all preemptible on the cluster.""")
@click.option(
    "--retries", type=int, help="""Specify the number of automatic retries for the experiment."""
)
def run(
    arg: Tuple[str, ...],
    name: Optional[str] = None,
    description: Optional[str] = None,
    task_name: str = "main",
    workspace: Optional[str] = None,
    cluster: Optional[Tuple[str, ...]] = None,
    hostname: Optional[Tuple[str, ...]] = None,
    beaker_image: Optional[str] = None,
    docker_image: Optional[str] = None,
    cpus: Optional[float] = None,
    gpus: Optional[int] = None,
    memory: Optional[str] = None,
    shared_memory: Optional[str] = None,
    dataset: Optional[Tuple[str, ...]] = None,
    gh_token_secret: str = constants.GITHUB_TOKEN_SECRET,
    ref: Optional[str] = None,
    conda: Optional[PathOrStr] = None,
    pip: Optional[PathOrStr] = None,
    venv: Optional[str] = None,
    env: Optional[Tuple[str, ...]] = None,
    env_secret: Optional[Tuple[str, ...]] = None,
    dataset_secret: Optional[Tuple[str, ...]] = None,
    timeout: int = 0,
    show_logs: bool = True,
    allow_dirty: bool = False,
    dry_run: bool = False,
    yes: bool = False,
    save_spec: Optional[PathOrStr] = None,
    priority: Optional[str] = None,
    install: Optional[str] = None,
    no_python: bool = False,
    no_conda: bool = False,
    replicas: Optional[int] = None,
    leader_selection: bool = False,
    host_networking: bool = False,
    propagate_failure: Optional[bool] = None,
    propagate_preemption: Optional[bool] = None,
    synchronized_start_timeout: Optional[str] = None,
    mount: Optional[Tuple[str, ...]] = None,
    weka: Optional[str] = None,
    budget: Optional[str] = None,
    preemptible: bool = False,
    stop_preemptible: bool = False,
    retries: Optional[int] = None,
):
    """
    Run an experiment on Beaker.

    Example:

    $ gantry run --name 'hello-world' -- python -c 'print("Hello, World!")'
    """
    if not arg:
        raise ConfigurationError(
            "[ARGS]... are required! For example:\n$ gantry run -- python -c 'print(\"Hello, World!\")'"
        )

    if beaker_image is None and docker_image is None:
        beaker_image = constants.DEFAULT_IMAGE
    elif (beaker_image is None) == (docker_image is None):
        raise ConfigurationError(
            "Either --beaker-image or --docker-image must be specified, but not both."
        )

    if budget is None:
        budget = prompt.Prompt.ask(
            "[yellow]Missing '--budget' option, "
            "see https://beaker-docs.apps.allenai.org/concept/budgets.html for more information.[/]\n"
            "[i]Please enter the budget account to associate with this experiment[/]",
        )

    if not budget:
        raise ConfigurationError("Budget account must be specified!")

    task_resources = TaskResources(
        cpu_count=cpus, gpu_count=gpus, memory=memory, shared_memory=shared_memory
    )

    # Get repository account, name, and current ref.
    github_account, github_repo, git_ref, is_public = util.ensure_repo(
        allow_dirty=ref is not None or allow_dirty
    )
    if ref is not None:
        git_ref = ref

    if not util.ref_exists_on_remote(git_ref):
        raise UnpushedChangesError(
            f"Current git ref '{git_ref}' does not appear to exist on the remote!\n"
            "Please push your changes and try again."
        )

    # Initialize Beaker client and validate workspace.
    beaker = util.ensure_workspace(
        workspace=workspace, yes=yes, gh_token_secret=gh_token_secret, public_repo=is_public
    )

    if beaker_image is not None and beaker_image != constants.DEFAULT_IMAGE:
        try:
            beaker_image = beaker.image.get(beaker_image).id
        except ImageNotFound:
            raise ConfigurationError(f"Beaker image '{beaker_image}' not found")

    # Get the entrypoint dataset.
    entrypoint_dataset = util.ensure_entrypoint_dataset(beaker)

    # Get / set the GitHub token secret.
    if not is_public:
        try:
            beaker.secret.get(gh_token_secret)
        except SecretNotFound:
            print_stderr(
                f"[yellow]GitHub token secret '{gh_token_secret}' not found in workspace.[/]\n"
                f"You can create a suitable GitHub token by going to https://github.com/settings/tokens/new "
                f"and generating a token with the '\N{ballot box with check} repo' scope."
            )
            gh_token = prompt.Prompt.ask(
                "[i]Please paste your GitHub token here[/]",
                password=True,
            )
            if not gh_token:
                raise ConfigurationError("token cannot be empty!")
            beaker.secret.write(gh_token_secret, gh_token)
            print(
                f"GitHub token secret uploaded to workspace as '{gh_token_secret}'.\n"
                f"If you need to update this secret in the future, use the command:\n"
                f"[i]$ gantry config set-gh-token[/]"
            )

        gh_token_secret = util.ensure_github_token_secret(beaker, gh_token_secret)

    # Validate the input datasets.
    datasets_to_use = ensure_datasets(beaker, *dataset) if dataset else []

    env_vars = []
    for e in env or []:
        try:
            env_name, val = e.split("=", 1)
        except ValueError:
            raise ValueError("Invalid --env option: {e}")
        env_vars.append((env_name, val))

    env_secrets = []
    for e in env_secret or []:
        try:
            env_secret_name, secret = e.split("=", 1)
        except ValueError:
            raise ValueError(f"Invalid --env-secret option: '{e}'")
        env_secrets.append((env_secret_name, secret))

    dataset_secrets = []
    for ds in dataset_secret or []:
        try:
            secret, mount_path = ds.split(":", 1)
        except ValueError:
            raise ValueError(f"Invalid --dataset-secret option: '{ds}'")
        dataset_secrets.append((secret, mount_path))

    mounts = []
    for m in mount or []:
        try:
            source, target = m.split(":", 1)
        except ValueError:
            raise ValueError(f"Invalid --mount option: '{m}'")
        mounts.append((source, target))

    weka_buckets = []
    for m in weka or []:
        try:
            source, target = m.split(":", 1)
        except ValueError:
            raise ValueError(f"Invalid --weka option: '{m}'")
        weka_buckets.append((source, target))

    # Validate clusters.
    if cluster:
        cl_objects = beaker.cluster.list()
        final_clusters = []
        for pat in cluster:
            matching_clusters = [cl.full_name for cl in cl_objects if fnmatch(cl.full_name, pat)]
            if matching_clusters:
                final_clusters.extend(matching_clusters)
            else:
                raise ConfigurationError(f"cluster '{pat}' did not match any Beaker clusters")
        cluster = list(set(final_clusters))  # type: ignore

    # Default to preemptible priority when no cluster has been specified.
    if not cluster and priority is None:
        priority = Priority.preemptible

    # Initialize experiment and task spec.
    spec = build_experiment_spec(
        task_name=task_name,
        clusters=list(cluster or []),
        task_resources=task_resources,
        arguments=list(arg),
        entrypoint_dataset=entrypoint_dataset.id,
        github_account=github_account,
        github_repo=github_repo,
        git_ref=git_ref,
        budget=budget,
        description=description,
        beaker_image=beaker_image,
        docker_image=docker_image,
        gh_token_secret=gh_token_secret if not is_public else None,
        conda=conda,
        pip=pip,
        venv=venv,
        datasets=datasets_to_use,
        env=env_vars,
        env_secrets=env_secrets,
        dataset_secrets=dataset_secrets,
        priority=priority,
        install=install,
        no_python=no_python,
        no_conda=no_conda,
        replicas=replicas,
        leader_selection=leader_selection,
        host_networking=host_networking or (bool(replicas) and leader_selection),
        propagate_failure=propagate_failure,
        propagate_preemption=propagate_preemption,
        synchronized_start_timeout=synchronized_start_timeout,
        mounts=mounts,
        weka_buckets=weka_buckets,
        hostnames=None if hostname is None else list(hostname),
        preemptible=preemptible,
        retries=retries,
    )

    if save_spec:
        if (
            Path(save_spec).is_file()
            and not yes
            and not prompt.Confirm.ask(
                f"[yellow]The file '{save_spec}' already exists. "
                f"[i]Are you sure you want to overwrite it?[/][/]"
            )
        ):
            raise KeyboardInterrupt
        spec.to_file(save_spec)
        print(f"Experiment spec saved to {save_spec}")

    if not name:
        default_name = util.unique_name()
        if yes:
            name = default_name
        else:
            name = prompt.Prompt.ask(
                "[i]What would you like to call this experiment?[/]", default=util.unique_name()
            )

    if not name:
        raise ConfigurationError("Experiment name cannot be empty!")

    if dry_run:
        rich.get_console().rule("[b]Dry run[/]")
        print(
            f"[b]Workspace:[/] {beaker.workspace.url()}\n"
            f"[b]Commit:[/] https://github.com/{github_account}/{github_repo}/commit/{git_ref}\n"
            f"[b]Name:[/] {name}\n"
            f"[b]Experiment spec:[/]",
            spec.to_json(),
        )
        return

    name_prefix = name
    while True:
        try:
            experiment = beaker.experiment.create(name, spec)
            break
        except ExperimentConflict:
            name = (
                name_prefix
                + "-"
                + random.choice(string.ascii_lowercase)
                + random.choice(string.ascii_lowercase)
                + random.choice(string.digits)
                + random.choice(string.digits)
            )

    print(f"Experiment submitted, see progress at {beaker.experiment.url(experiment)}")

    if stop_preemptible:
        if priority == Priority.preemptible:
            print_stderr("[yellow]You cannot preempt other jobs when your job is preemptible.[/]")
        elif not cluster:
            print_stderr("[yellow]Preempt jobs requires specifying a cluster.[/]")
        elif len(cluster) > 1:
            print_stderr("[yellow]Preempt jobs requires specifying a single cluster.[/]")
        elif not dry_run:
            print(f"Preempting jobs on cluster {cluster[0]}...")
            preempted = beaker.cluster.preempt_jobs(cluster[0], ignore_failures=True)
            if preempted:
                print(f"Preempted {len(preempted)} jobs on cluster {cluster[0]}")
            else:
                print("No more jobs to preempt")

    # Can return right away if timeout is 0.
    if timeout == 0:
        return

    job: Optional[Job] = None
    try:
        if show_logs:
            job = util.follow_experiment(beaker, experiment, timeout=timeout)
        else:
            experiment = beaker.experiment.wait_for(
                experiment, timeout=timeout if timeout > 0 else None
            )[0]
            job = beaker.experiment.tasks(experiment)[0].latest_job  # type: ignore
            assert job is not None
    except (TermInterrupt, JobTimeoutError) as exc:
        print_stderr(f"[red][bold]{exc.__class__.__name__}:[/] [i]{exc}[/][/]")
        beaker.experiment.stop(experiment)
        print_stderr("[yellow]Experiment cancelled.[/]")
        sys.exit(1)
    except KeyboardInterrupt as exc:
        print_stderr(f"[red][bold]{exc.__class__.__name__}:[/] [i]{exc}[/][/]")
        print(f"See the experiment at {beaker.experiment.url(experiment)}")
        print_stderr(
            f"[yellow]To cancel the experiment, run:\n[i]$ gantry stop {experiment.id}[/][/]"
        )
        sys.exit(1)

    util.display_results(beaker, experiment, job)


def build_experiment_spec(
    *,
    task_name: str,
    clusters: List[str],
    task_resources: TaskResources,
    arguments: List[str],
    entrypoint_dataset: str,
    github_account: str,
    github_repo: str,
    git_ref: str,
    budget: str,
    description: Optional[str] = None,
    beaker_image: Optional[str] = None,
    docker_image: Optional[str] = None,
    gh_token_secret: Optional[str] = constants.GITHUB_TOKEN_SECRET,
    conda: Optional[PathOrStr] = None,
    pip: Optional[PathOrStr] = None,
    venv: Optional[str] = None,
    datasets: Optional[List[Tuple[str, Optional[str], str]]] = None,
    env: Optional[List[Tuple[str, str]]] = None,
    env_secrets: Optional[List[Tuple[str, str]]] = None,
    dataset_secrets: Optional[List[Tuple[str, str]]] = None,
    priority: Optional[Union[str, Priority]] = None,
    install: Optional[str] = None,
    no_python: bool = False,
    no_conda: bool = False,
    replicas: Optional[int] = None,
    leader_selection: bool = False,
    host_networking: bool = False,
    propagate_failure: Optional[bool] = None,
    propagate_preemption: Optional[bool] = None,
    synchronized_start_timeout: Optional[str] = None,
    mounts: Optional[List[Tuple[str, str]]] = None,
    weka_buckets: Optional[List[Tuple[str, str]]] = None,
    hostnames: Optional[List[str]] = None,
    preemptible: bool = False,
    retries: Optional[int] = None,
):
    task_spec = (
        TaskSpec.new(
            task_name,
            beaker_image=beaker_image,
            docker_image=docker_image,
            result_path=constants.RESULTS_DIR,
            command=["bash", "/gantry/entrypoint.sh"],
            arguments=arguments,
            resources=task_resources,
            priority=priority,
            replicas=replicas,
            leader_selection=leader_selection,
            host_networking=host_networking,
            propagate_failure=propagate_failure,
            propagate_preemption=propagate_preemption,
            synchronized_start_timeout=synchronized_start_timeout,
        )
        .with_env_var(name="GANTRY_VERSION", value=VERSION)
        .with_env_var(name="GITHUB_REPO", value=f"{github_account}/{github_repo}")
        .with_env_var(name="GIT_REF", value=git_ref)
        .with_env_var(name="GANTRY_TASK_NAME", value=task_name)
        .with_dataset("/gantry", beaker=entrypoint_dataset)
    )

    if preemptible:
        task_spec.context.preemptible = True

    if clusters:
        task_spec = task_spec.with_constraint(cluster=clusters)

    if hostnames:
        task_spec = task_spec.with_constraint(hostname=hostnames)

    if gh_token_secret is not None:
        task_spec = task_spec.with_env_var(name="GITHUB_TOKEN", secret=gh_token_secret)

    for name, val in env or []:
        task_spec = task_spec.with_env_var(name=name, value=val)

    for name, secret in env_secrets or []:
        task_spec = task_spec.with_env_var(name=name, secret=secret)

    if no_python:
        task_spec = task_spec.with_env_var(name="NO_PYTHON", value="1")
    else:
        if not no_conda:
            if conda is not None:
                task_spec = task_spec.with_env_var(
                    name="CONDA_ENV_FILE",
                    value=str(conda),
                )
            elif Path(constants.CONDA_ENV_FILE).is_file():
                task_spec = task_spec.with_env_var(
                    name="CONDA_ENV_FILE",
                    value=constants.CONDA_ENV_FILE,
                )
            elif Path(constants.CONDA_ENV_FILE_ALTERNATE).is_file():
                task_spec = task_spec.with_env_var(
                    name="CONDA_ENV_FILE",
                    value=constants.CONDA_ENV_FILE_ALTERNATE,
                )
            else:
                task_spec = task_spec.with_env_var(
                    name="PYTHON_VERSION", value=".".join(platform.python_version_tuple()[:-1])
                )

            if venv is not None:
                task_spec = task_spec.with_env_var(
                    name="VENV_NAME",
                    value=venv,
                )
        else:
            task_spec = task_spec.with_env_var(name="NO_CONDA", value="1")

        if pip is not None:
            task_spec = task_spec.with_env_var(
                name="PIP_REQUIREMENTS_FILE",
                value=str(pip),
            )

        if install is not None:
            task_spec = task_spec.with_env_var(name="INSTALL_CMD", value=install)

    if datasets:
        for dataset_id, sub_path, path in datasets:
            task_spec = task_spec.with_dataset(path, beaker=dataset_id, sub_path=sub_path)

    for secret, mount_path in dataset_secrets or []:
        task_spec = task_spec.with_dataset(mount_path, secret=secret)

    if mounts:
        for source, target in mounts:
            task_spec = task_spec.with_dataset(target, host_path=source)

    if weka_buckets:
        for source, target in weka_buckets:
            task_spec = task_spec.with_dataset(target, weka=source)

    return ExperimentSpec(
        description=description,
        budget=budget,
        tasks=[task_spec],
        retry=None if not retries else RetrySpec(allowed_task_retries=retries),
    )


def ensure_datasets(beaker: Beaker, *datasets: str) -> List[Tuple[str, Optional[str], str]]:
    out = []
    for dataset_str in datasets:
        dataset_name: str
        path: str
        sub_path: Optional[str] = None
        if dataset_str.count(":") == 1:
            dataset_name, path = dataset_str.split(":")
        elif dataset_str.count(":") == 2:
            dataset_name, sub_path, path = dataset_str.split(":")
        else:
            raise ValueError(
                f"Bad '--dataset' specification: '{dataset_str}'\n"
                f"Datasets should be in the form of 'dataset-name:/mount/location'"
                f"or 'dataset-name:sub/path:/mount/location'"
            )
        dataset_id = beaker.dataset.get(dataset_name).id
        out.append((dataset_id, sub_path, path))
    return out



from datetime import datetime, timedelta, timezone
from itertools import islice
from typing import Generator, List, Optional, Tuple

import click
from beaker import Beaker, Experiment, Task, Tasks
from rich import print
from rich.progress import Progress
from rich.table import Table

from ..exceptions import ConfigurationError
from ..util import StrEnum
from .main import CLICK_COMMAND_DEFAULTS, main


class JobStatus(StrEnum):
    running = "running"
    created = "created"
    scheduled = "scheduled"
    failed = "failed"
    canceled = "canceled"
    succeeded = "succeeded"


class Defaults:
    limit = 10
    max_age = 7


@main.command(name="list", **CLICK_COMMAND_DEFAULTS)
@click.option(
    "-w",
    "--workspace",
    type=str,
    help="""The Beaker workspace to pull experiments from.
    If not specified, your default workspace will be used.""",
)
@click.option(
    "-l",
    "--limit",
    type=int,
    default=Defaults.limit,
    help=f"Limit the number of experiments to display. Default: {Defaults.limit}",
)
@click.option(
    "-a",
    "--author",
    type=str,
    help="Filter by author. Tip: use '--me' instead to show your own experiments.",
)
@click.option(
    "--me",
    is_flag=True,
    help="Only show your own experiments. Mutually exclusive with '--author'.",
)
@click.option(
    "-s",
    "--status",
    type=click.Choice(list(JobStatus)),
    help="Filter by status.",
    multiple=True,
)
@click.option(
    "--max-age",
    type=int,
    default=Defaults.max_age,
    help=f"Maximum age of experiments, in days. Default: {Defaults.max_age}",
)
def list_cmd(
    workspace: Optional[str] = None,
    limit: int = Defaults.limit,
    author: Optional[str] = None,
    me: bool = False,
    status: Optional[List[str]] = None,
    max_age: int = Defaults.max_age,
):
    """
    List gantry experiments.
    """
    beaker = Beaker.from_env(session=True)

    table = Table(title="Experiments", show_lines=True)
    table.add_column("Name", justify="left", no_wrap=True)
    table.add_column("Author", justify="left", style="blue", no_wrap=True)
    table.add_column("Created", justify="left", no_wrap=True)
    table.add_column("Tasks")

    if me:
        if author is not None:
            raise ConfigurationError("--me and -a/--author are mutually exclusive.")
        else:
            author = beaker.account.whoami().name

    with Progress(transient=True) as progress:
        task = progress.add_task("Collecting experiments...", total=limit)
        for exp, tasks in islice(
            iter_experiments(
                beaker, workspace=workspace, author=author, statuses=status, max_age=max_age
            ),
            limit,
        ):
            table.add_row(
                f"[b cyan]{exp.display_name}[/]\n[u i blue]{beaker.experiment.url(exp)}[/]",
                exp.author.name,
                exp.created.astimezone(tz=None).strftime("%I:%M %p on %a, %b %-d"),
                "\n".join(format_task(task) for task in tasks),
            )
            progress.update(task, advance=1)

        progress.update(task, completed=True)

    print(table)


def iter_experiments(
    beaker: Beaker,
    *,
    workspace: Optional[str],
    author: Optional[str],
    statuses: Optional[List[str]],
    max_age: int,
) -> Generator[Tuple[Experiment, Tasks], None, None]:
    now = datetime.now(tz=timezone.utc).astimezone()
    for exp in beaker.workspace.iter_experiments(workspace=workspace):
        # Filter by age.
        age = now - exp.created.astimezone()
        if age > timedelta(days=max_age):
            break

        # Maybe filter by author.
        if author is not None and exp.author.name != author:
            continue

        # Filter out non-gantry experiments.
        spec = beaker.experiment.spec(exp)
        for task_spec in spec.tasks:
            for env_var in task_spec.env_vars or []:
                if env_var.name == "GANTRY_VERSION":
                    break
            else:
                continue

        tasks = beaker.experiment.tasks(exp)

        # Maybe filter by status.
        if statuses:
            for task in tasks:
                if get_status(task) in statuses:
                    break
            else:
                continue

        yield exp, tasks


def get_status(task: Task) -> Optional[JobStatus]:
    job = task.latest_job
    status: Optional[JobStatus] = None
    if job is not None:
        exit_code = job.status.exit_code
        if job.status.current in ("running", "created", "scheduled"):
            status = JobStatus(job.status.current)
        elif job.status.failed is not None or (exit_code is not None and exit_code > 0):
            status = JobStatus.failed
        elif job.status.canceled is not None:
            status = JobStatus.canceled
        elif job.status.finalized is not None and exit_code == 0:
            status = JobStatus.succeeded
    return status


def format_task(task: Task) -> str:
    style = "i"
    status = get_status(task)
    if status in (JobStatus.running, JobStatus.succeeded):
        style += " green"
    elif status in (JobStatus.scheduled, JobStatus.created, JobStatus.canceled):
        style += " yellow"
    elif status == JobStatus.failed:
        style += " red"
    return f"[i]{task.display_name}[/] ([{style}]{status or 'unknown'}[/])"



from .cluster import cluster
from .completion import completion
from .config import config
from .follow import follow
from .list import list_cmd
from .logs import logs
from .main import main
from .run import run
from .stop import stop

__all__ = ["main", "follow", "run", "config", "cluster", "stop", "list_cmd", "completion", "logs"]
